/*
	(C) 1995-2000 AROS - The Amiga Research OS
	$Id: protect.c,v 1.3 2021/01/20 11:57:56 piru Exp $
 
	Desc: Protect CLI command
	Lang: English
*/

/*****************************************************************************
 
	NAME
 
		Protect
 
	SYNOPSIS
 
		FILE/A,FLAGS,ADD/S,SUB/S,ALL/S,QUIET/S
 
	LOCATION
 
		Workbench:c
 
	FUNCTION
 
		Add or remove protection bits from a file or directory.
		
		Protect allows the use of pattern matching and recursive directory
		scans to protect many files/directories at any one time.
 
	INPUTS
 
		FILE   --  Either a file, a directory or a pattern to match.
		FLAGS  --  One or more of the following flags:
 
					S - Script
					P - Pure
					A - Archive
					R - Read
					W - Write
					E - Execute
					D - Delete
 
		ADD    --  Allows the bits to be set and hence allowable.
		SUB    --  Allows the bits to be cleared and hence not allowable.
		ALL    --  Allows a recursive scan of the volume/directory.
		QUIET  --  Suppresses any output to the shell.
 
	RESULT
 
		Standard DOS return codes.
 
	NOTES
 
	EXAMPLE
 
		Protect ram: e add all
 
			Recurses the ram: volume and attaches the executable bit.
 
	BUGS
 
	SEE ALSO
 
		SetProtection()
 
	INTERNALS
 
	HISTORY
 
		27-Jul-1997  laguest  --  Initial inclusion into the AROS tree
	3.12.2000    SDuvan   --  Rewrote, simplified and implemented missing
	                          functionality
 
******************************************************************************/

#define USE_SOFTLINKCHECK		1
#define USE_KEEPOTHERFLAGS		1

#include <dos/dos.h>
#include <dos/dosasl.h>
#include <dos/dosextens.h>
#include <dos/exall.h>
#include <dos/rdargs.h>
#include <exec/memory.h>
#include <exec/types.h>
#include <utility/utility.h>

//#include <proto/arossupport.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/utility.h>

#include <ctype.h>
#include <stdio.h>

#define	printf	Printf


#define MAX_PATH_LEN		512



#define Bit_Mask(bit)           (1L << bit)
#define Bit_Clear(name, bit)    name &= ~Bit_Mask(bit)
#define Bit_Set(name, bit)      name |= Bit_Mask(bit)

#define ARG_TEMPLATE    "FILE/A,FLAGS,ADD/S,SUB/S,ALL/S,QUIET/S"

#ifndef FIBB_HOLD
#define FIBB_HOLD 7
#endif
#ifndef FIBF_HOLD
#define FIBF_HOLD (1<<FIBB_HOLD)
#endif

enum
{
    ARG_FILE = 0,
    ARG_FLAGS,
    ARG_ADD,
    ARG_SUB,
    ARG_ALL,
    ARG_QUIET,
    NOOFARGS
};


/* To define whether a command line switch was set or not.
 */
#define NOT_SET(x)      (x == 0)
#define IS_SET(x)       (!NOT_SET(x))

#include "protect_version.h"

static const char version[] = VERSTAG;



struct data
{
    struct ExecBase	*SysBase;
    struct DosLibrary	*DOSBase;
    struct Library	*UtilityBase;
    CONST_STRPTR         errheader;
    LONG                 ioerr;
};

#define SysBase		data->SysBase
#define DOSBase		data->DOSBase
#define UtilityBase	data->UtilityBase
#define g_errheader	data->errheader
#define g_ioerr		data->ioerr


#if 0
static
int Do_Protect(struct AnchorPath *, STRPTR, STRPTR, BOOL, BOOL, BOOL, BOOL, struct data *);
#endif

static
int doProtect(struct AnchorPath *ap, STRPTR file, LONG flags, BOOL flagsSet,
              BOOL add, BOOL sub, BOOL all, BOOL quiet, struct data *);

static
BOOL setProtection(STRPTR file, LONG oldFlags, LONG flags, BOOL flagsSet,
                   BOOL add, BOOL sub, struct data *);

int main(void)
{
    struct data _data, *data = &_data;
    struct RDArgs     *rda;
    struct AnchorPath *apath;

    IPTR args[NOOFARGS] = { NULL,
                            NULL,
                            (IPTR)FALSE,
                            (IPTR)FALSE,
                            (IPTR)FALSE,
                            (IPTR)FALSE };

    int retval = RETURN_FAIL;

    SysBase=*((struct ExecBase**) 4);
    DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37);
    UtilityBase=OpenLibrary("utility.library",37);
    if (DOSBase && UtilityBase)
    {
        g_errheader = NULL;
        g_ioerr     = 0;

        apath = AllocVec(sizeof(struct AnchorPath) + MAX_PATH_LEN,
                         MEMF_ANY | MEMF_CLEAR);

        if (apath)
        {
            retval = RETURN_ERROR;

            rda = ReadArgs(ARG_TEMPLATE, args, NULL);

            if (rda != NULL)
            {
                STRPTR  file = (STRPTR)args[ARG_FILE];
                STRPTR  flags = (STRPTR)args[ARG_FLAGS];
                BOOL    add = (BOOL)args[ARG_ADD];
                BOOL    sub = (BOOL)args[ARG_SUB];
                BOOL    all = (BOOL)args[ARG_ALL];
                BOOL    quiet = (BOOL)args[ARG_QUIET];

                LONG    flagValues = FIBF_READ | FIBF_WRITE | FIBF_DELETE |
                                     FIBF_EXECUTE;

                retval = RETURN_OK;

                if (flags != NULL)
                {
                    if (*flags == '+')
                    {
                        add = TRUE;
                        flags++;
                    }
                    if (*flags == '-')
                    {
                        sub = TRUE;
                        flags++;
                    }

                    while (*flags != 0 && retval == RETURN_OK)
                    {
                        switch (ToUpper(*flags))
                        {
                            /* Active low */
                        case 'R':
                            flagValues &= ~FIBF_READ;
                            break;

                        case 'W':
                            flagValues &= ~FIBF_WRITE;
                            break;

                        case 'D':
                            flagValues &= ~FIBF_DELETE;
                            break;

                        case 'E':
                            flagValues &= ~FIBF_EXECUTE;
                            break;

                            /* Active high */
                        case 'A':
                            flagValues |= FIBF_ARCHIVE;
                            break;

                        case 'S':
                            flagValues |= FIBF_SCRIPT;
                            break;

                        case 'P':
                            flagValues |= FIBF_PURE;
                            break;

                        case 'H':
                            flagValues |= FIBF_HOLD;
                            break;

                        default:
                            retval = RETURN_FAIL;
                            break;
                        }

                        flags++;
                    } /* while (*flags != 0) */
                }
                else
                {
                    retval = RETURN_FAIL;
                }

                if (retval)
                {
                    g_errheader = "Invalid flag - must be one of HSPARWED\n";
                }
                else if (add && sub)
                {
                    g_errheader = "Can't specify both ADD (+) and SUB (-)\n";
                    retval = RETURN_FAIL;
                }

                if (retval == RETURN_OK)
                {
                    if (!all && IsDosEntryA(file, LDF_VOLUMES | LDF_DEVICES, DOSBase, UtilityBase))
                    {
                        stccpy(apath->ap_Buf, file, MAX_PATH_LEN);
                        g_errheader = "Can't set protection for %s - ";
                        g_ioerr = ERROR_OBJECT_WRONG_TYPE;

                        retval = RETURN_FAIL;
                    }
                    else
                    {
                        retval = doProtect(apath, file, flagValues,
                                           flags != NULL, add, sub, all,
                                           quiet, data);
                    }
                }

                FreeArgs(rda);
            }
            else
            {
                g_ioerr = IoErr();
                retval = RETURN_FAIL;
            }
        }
        else
        {
            g_ioerr = IoErr();
            retval = RETURN_FAIL;
        }

        if (g_errheader)
        {
            printf(g_errheader, (IPTR) apath->ap_Buf);
        }

        FreeVec(apath);

        if (g_ioerr)
        {
            PrintFault(g_ioerr, NULL);
            SetIoErr(g_ioerr);
        }
    }

    if (UtilityBase)
        CloseLibrary(UtilityBase);

    if (DOSBase)
        CloseLibrary((struct Library *)DOSBase);

    return retval;
}


/* Defines whether MatchFirst(), etc has matched a file. */
#define MATCHED_FILE    0

#if 0
static
int Do_Recursion(struct AnchorPath *, LONG, LONG, LONG *, LONG, struct data *);
static
void AddBitMask(struct AnchorPath *, STRPTR, LONG *, BOOL);
static
void SubBitMask(struct AnchorPath *, STRPTR, LONG *, BOOL);
static
void NewBitMask(struct AnchorPath *, STRPTR, LONG *);
#endif

#if 0
static
void PrintFileName(struct AnchorPath *, LONG, struct data *);
static
int SafeSetProtection(struct AnchorPath *, LONG, struct data *);
#endif


#define  isDir(fib) ((fib)->fib_DirEntryType >= 0)

static
int doProtect(struct AnchorPath *ap, STRPTR file, LONG flags, BOOL flagsSet,
              BOOL add, BOOL sub, BOOL all, BOOL quiet, struct data *data)
{
    LONG  match;
    int   retval = RETURN_OK;
    LONG  indent = 0;
    int   i;			/* Loop variable */
    BOOL  success;

    /* Make sure DOS knows the buffer size. */
    ap->ap_Flags      = 0; // APF_DOWILD
    ap->ap_Strlen     = MAX_PATH_LEN;
    ap->ap_BreakBits  = SIGBREAKF_CTRL_C;
    ap->ap_FoundBreak = 0;

    for (match = MatchFirst(file, ap); match == 0;
         match = MatchNext(ap))
    {
        if (isDir(&ap->ap_Info))
        {
            if (ap->ap_Flags & APF_DIDDIR)
            {
                indent--;
                ap->ap_Flags &= ~APF_DIDDIR; /* Should not be necessary */
                continue;
            }
            else if (all)
            {

#if USE_SOFTLINKCHECK

                if (SoftlinkDODIR(ap, TRUE, quiet, DOSBase))
                {
                    ap->ap_Flags |= APF_DODIR;

                    indent++;
                }

#else /* USE_SOFTLINKCHECK */

                ap->ap_Flags |= APF_DODIR;

                indent++;

#endif /* USE_SOFTLINKCHECK */

            }
        }

        success = setProtection(ap->ap_Buf, ap->ap_Info.fib_Protection, flags,
                                flagsSet, add, sub, data);

        if (!success)
        {
            g_ioerr = IoErr();
        }

        if (!quiet && all)
        {
            /* Fix indentation level */
            for (i = 0; i < indent; i++)
            {
                PutStr("     ");
            }

            if (!isDir(&ap->ap_Info))
            {
                PutStr("   ");
            }

            PutStr(ap->ap_Info.fib_FileName);

            if (isDir(&ap->ap_Info))
            {
                PutStr(" (dir)");
            }

            PutStr("..");

            if (success)
            {
                PutStr("done\n");
            }
        }

        if (!success)
        {
            g_errheader = "Can't set protection for %s - ";
            match = g_ioerr;
            break;
        }
    }

    MatchEnd(ap);

    if (match == ERROR_NO_MORE_ENTRIES)
    {
      retval = RETURN_OK;
    }
    else if (match == ERROR_BREAK)
    {
      g_ioerr = match;
      retval = RETURN_WARN;
    }
    else
    {
      g_ioerr = match;
      retval = RETURN_FAIL;
    }

    return retval;
}

#define  ALL_OFF  (FIBF_READ | FIBF_WRITE | FIBF_DELETE | FIBF_EXECUTE)
#define  addFlags(new, old)  ((~(~old | ~new) & ALL_OFF) | \
((old | new) & ~ALL_OFF))

#define  subFlags(new, old)  (((old | ~new) & ALL_OFF) | \
((old & ~new) & ~ALL_OFF))

static
BOOL setProtection(STRPTR file, LONG oldFlags, LONG flags, BOOL flagsSet,
                   BOOL add, BOOL sub, struct data *data)
{
    LONG  newFlags;

    if (flags != ALL_OFF)
    {
        if (add)
        {
            /* Enable permission */
            newFlags = addFlags(flags, oldFlags);
        }
        else if (sub)
        {
            /* Disable permissions */
            newFlags = subFlags(flags, oldFlags);
        }
        else
        {
#if USE_KEEPOTHERFLAGS
            /* Clear all permissions then set the ones given. Keep the other bits, however.
               This is what Protect 42.x does. - Piru */
            newFlags = (oldFlags & ~0xff) | flags;
#else
            /* Clear all permissions then set the ones given. */
            newFlags = flags;
#endif /* USE_KEEPOTHERFLAGS */
        }
    }
    else
    {
        /* No flags were given */
        if (!add && !sub)
        {
            /* Disable all permissions */
            newFlags = ALL_OFF;
        }
        else
        {
            /* Do nothing */
            return FALSE;
        }
    }

    if (newFlags != oldFlags && !SetProtection(file, newFlags))
    {
        return FALSE;
    }

    return TRUE;
}


#if 0
static
int Do_Protect(struct AnchorPath *a, STRPTR file, STRPTR flags,
               BOOL add, BOOL sub, BOOL all, BOOL quiet, struct data *data)
{
    LONG Result,
    TabValue,
    BitMask;
    int  retval = RETURN_OK;
    IPTR DevArg[2];

    TabValue     = 0L;
    BitMask      = 0L;

    if(IsDosEntryA(file, LDF_VOLUMES | LDF_DEVICES, DOSBase, UtilityBase) && !all)
    {
        DevArg[0] = (IPTR)file;

        VPrintf("Can't set protection for %s - ", DevArg);
        SetIoErr(ERROR_OBJECT_WRONG_TYPE);
        PrintFault(IoErr(), NULL);

        retval = RETURN_FAIL;
    }
    else
    {
        Result = MatchFirst(file, a);

        if (Result == MATCHED_FILE)
        {
            do
            {
                if (flags != NULL)  /* Flags are given */
                {
                    if (add)
                    {
                        /* Add the permission bits to the bitmask.
                         * In other words - Enable permissions.
                         */

                        AddBitMask(a, flags, &BitMask, TRUE);
                        retval = Do_Recursion(a, all, quiet, &TabValue, BitMask, data);
                    }
                    else if (sub)
                    {
                        /* Remove the permission bits from the bitmask.
                         * In other words - Disable permissions. */

                        SubBitMask(a, flags, &BitMask, TRUE);
                        retval = Do_Recursion(a, all, quiet, &TabValue, BitMask, data);
                    }
                    else
                    {
                        /* Clear all permissions then set the ones given.
                         */

                        NewBitMask(a, flags, &BitMask);
                        retval = Do_Recursion(a, all, quiet, &TabValue, BitMask, data);
                    }
                }
                else    /* No flags are given */
                {
                    if (!add && !sub)
                    {
                        /* Remove all permissions bits from the bitmask.
                         * In other words - Disable ALL permissions. */

                        SubBitMask(a, "sparwed", &BitMask, TRUE);
                        retval = Do_Recursion(a, all, quiet, &TabValue, BitMask, data);
                    }
                }
            }

            while (((Result = MatchNext(a)) == MATCHED_FILE) &&
                    retval != RETURN_FAIL);
        }
        else
        {
            PrintFault(IoErr(), "Protect");
            retval = RETURN_FAIL;
        }

        MatchEnd(a);
    }

    return retval;
}
#endif


#if 0
static
int Do_Recursion(struct AnchorPath *a,
                 LONG  All,
                 LONG  Quiet,
                 LONG *TabValue,
                 LONG  BitMask, struct data *data)
{
    BOOL Return_Value;

    Return_Value = RETURN_OK;

    if (NOT_SET(All))
    {
        Return_Value = SafeSetProtection(a, BitMask, data);
        if (NOT_SET(Quiet))
        {
            PrintFileName(a, *TabValue, data);
        }
    }
    else
    {
        /* Allow a recursive scan.
         */

        if (a->ap_Info.fib_DirEntryType > 0)
        {
            /* Enter directory.
             */
            if (!(a->ap_Flags & APF_DIDDIR))
            {
                a->ap_Flags |= APF_DODIR;

                Return_Value = SafeSetProtection(a, BitMask, data);

                if (NOT_SET(Quiet))
                {
                    PrintFileName(a, *TabValue, data);
                }
                (*TabValue)++;
            }
            else
            {
                /* Leave directory.
                 */

                a->ap_Flags &= ~APF_DIDDIR;
                (*TabValue)--;
            }
        }
        else
        {
            Return_Value = SafeSetProtection(a, BitMask, data);
            if (NOT_SET(Quiet))
            {
                PrintFileName(a, *TabValue, data);
            }
        }
    }

    return (Return_Value);

} /* Do_Recursion */
#endif


#if 0
/* AddBitMask
 *
 * Create a bitmask that unsets the appropriate bits.
 */
static
void AddBitMask(struct AnchorPath *a, STRPTR f, LONG *b, BOOL copy)
{
    int LoopCount;

    LoopCount = 0;

    if (copy == TRUE)
    {
        *b        = a->ap_Info.fib_Protection;
    }

    while (f[LoopCount] != NULL)
    {
        /* The SetProtect() function uses a clear bit to denote
         * that one of the RWED permissions are allowed.
         *
         * It also uses a set bit to denote that one of the
         * SPA permissions are allowed.
         */

        switch (f[LoopCount])
        {
        case 'h':   /* Hold */
            Bit_Set(*b, FIBB_HOLD);
            break;
        case 's':   /* Script */
            Bit_Set(*b, FIBB_SCRIPT);
            break;
        case 'p':   /* Pure */
            Bit_Set(*b, FIBB_PURE);
            break;
        case 'a':   /* Archive */
            Bit_Set(*b, FIBB_ARCHIVE);
            break;
        case 'r':   /* Read */
            Bit_Clear(*b, FIBB_READ);
            break;
        case 'w':   /* Write */
            Bit_Clear(*b, FIBB_WRITE);
            break;
        case 'e':   /* Execute */
            Bit_Clear(*b, FIBB_EXECUTE);
            break;
        case 'd':   /* Delete */
            Bit_Clear(*b, FIBB_DELETE);
            break;
        default:
            break;
        }

        LoopCount++;
    }
} /* AddBitMask */


/* SubBitMask
 *
 * Create a bitmask that sets the appropriate bits.
 */
static
void SubBitMask(struct AnchorPath *a, STRPTR f, LONG *b, BOOL copy)
{
    int LoopCount;

    LoopCount = 0;

    if (copy == TRUE)
    {
        *b        = a->ap_Info.fib_Protection;
    }

    while (f[LoopCount] != NULL)
    {
        /* The SetProtect() function uses a set bit to denote
         * that one of the RWED permissions is not allowed.
         *
         * It also uses a set bit to denote that one of the
         * SPA permissions is not allowed.
         */

        switch (f[LoopCount])
        {
        case 'h':   /* Hold */
            Bit_Clear(*b, FIBB_HOLD);
            break;
        case 's':   /* Script */
            Bit_Clear(*b, FIBB_SCRIPT);
            break;
        case 'p':   /* Pure */
            Bit_Clear(*b, FIBB_PURE);
            break;
        case 'a':   /* Archive */
            Bit_Clear(*b, FIBB_ARCHIVE);
            break;
        case 'r':   /* Read */
            Bit_Set(*b, FIBB_READ);
            break;
        case 'w':   /* Write */
            Bit_Set(*b, FIBB_WRITE);
            break;
        case 'e':   /* Execute */
            Bit_Set(*b, FIBB_EXECUTE);
            break;
        case 'd':   /* Delete */
            Bit_Set(*b, FIBB_DELETE);
            break;
        default:
            break;
        }

        LoopCount++;
    }
} /* SubBitMask */


/* NewBitMask
 *
 * Create a new bitmask that has the appropriate bits set.
 */
static
void NewBitMask(struct AnchorPath *a, STRPTR f, LONG *b)
{
    int LoopCount;

    LoopCount = 0;

    SubBitMask(a, "hsparwed", b, TRUE);
    AddBitMask(a, f, b, FALSE);
} /* NewBitMask */
#endif


#if 0
static
void PrintFileName(struct AnchorPath *a, LONG t, struct data *data)
{
    int i;

    for (i = 0; i < t; i++)
    {
        PutStr("     ");
    }

    if (a->ap_Info.fib_DirEntryType <= 0)
    {
        PutStr("   ");
    }

    PutStr(a->ap_Info.fib_FileName);

    if (a->ap_Info.fib_DirEntryType > 0)
    {
        PutStr(" (dir)");
    }

    PutStr("..done\n");

} /* PrintFileName */
#endif


#if 0
static
int SafeSetProtection(struct AnchorPath *a, LONG b, struct data *data)
{
    int Return_Value;

    Return_Value = RETURN_OK;

    if (!SetProtection(a->ap_Buf, b))
    {
        Return_Value = RETURN_WARN;
    }

    return(Return_Value);

} /* SafeSetProtection */
#endif
